from __future__ import annotations

import datetime
from typing import List, Optional, Sequence, Union

from anchorpoint import TextPositionSelector, TextQuoteSelector, TextPositionSet
from anchorpoint.schemas import TextPositionSetFactory
from pydantic import BaseModel, HttpUrl, validator

from justopinion.citations import CAPCitation


class ReporterVolume(BaseModel):
    """
    A group of decisions corresponding to a bound print volume.

    :param id:
        The unique identifier for this volume.
    :param url:
        The URL of this volume in the Case Access Project API.
    :param full_name:
        The name of this volume.
    """

    id: int
    url: HttpUrl
    full_name: str


class Court(BaseModel):
    """
    A court that issues legal decisions.

    :param id:
        The unique identifier for this court.
    :param name:
        The name of this court.
    :param url:
        The URL of this court in the Case Access Project API.
    :param slug:
        The URL-safe name of this court.
    :param name_abbreviation:
        The abbreviation of this court's name.
    """

    id: int
    name: str
    url: HttpUrl
    slug: str
    name_abbreviation: str = ""


class Jurisdiction(BaseModel):
    """
    A government or other entity that is responsible for a legal system.

    :param id:
        The unique identifier for this jurisdiction.
    :param name:
        The name of this jurisdiction.
    :param url:
        The URL of this jurisdiction in the Case Access Project API.
    :param slug:
        The URL-safe name of this jurisdiction.
    :param whitelisted:
        Whether the jurisdiction's cases are accessible without restriction
        in the Case Access Project API.
    :param name_abbreviation:
        The abbreviation of this jurisdiction.
    """

    id: int
    name: str
    url: HttpUrl
    slug: str
    whitelisted: bool
    name_abbreviation: str = ""


class Opinion(BaseModel):
    """
    A document that resolves legal issues in a case and posits legal holdings.

    Usually an opinion must have ``type="majority"``
    to create holdings binding on any courts.

    :param type:
        the opinion's attitude toward the court's disposition of the case.
        e.g. ``majority``, ``dissenting``, ``concurring``, ``concurring in the result``
    :param author:
        name of the judge who authored the opinion, if identified
    :param text:
    """

    type: str = "majority"
    author: str = ""
    text: str = ""

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(type="{self.type}", author="{self.author}")'

    def __str__(self):
        result = f"{self.type} opinion" if self.type else "opinion"
        if self.author:
            result += f" by {self.author}"
        return result

    @validator("author", pre=True)
    def remove_author_name_punctuation(cls, v):
        result = v or ""
        result = result.replace("Judge.", "Judge").replace("Justice.", "Justice")
        return result.strip(", -:;")

    def locate_text(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextQuoteSelector,
            Sequence[Union[str, TextQuoteSelector, TextPositionSelector]],
        ],
    ) -> TextPositionSet:
        """Get set of position selectors for text in Opinion."""
        factory = TextPositionSetFactory(self.text)
        return factory.from_selection(selection)

    def select_text(self, selector: TextQuoteSelector) -> Optional[str]:
        r"""
        Get text using a :class:`.TextQuoteSelector`.

        :param selector:
            a selector referencing a text passage in the :class:`Opinion`.

        :returns:
            the text referenced by the selector, or ``None`` if the text
            can't be found.
        """
        text_locations = self.locate_text(selector)
        return text_locations.as_text_sequence(self.text)


class CaseData(BaseModel):
    """
    The content of a Decision, including Opinions.
    """

    head_matter: Optional[str] = None
    corrections: Optional[str] = None
    parties: List[str] = []
    attorneys: List[str] = []
    opinions: List[Opinion] = []
    judges: List[str] = []


class PageRank(BaseModel):
    """
    The rank of the Decision in the collection of documents.
    """

    percentile: float
    raw: float


class CaseBody(BaseModel):

    data: CaseData
    status: str = ""


class DecisionAnalysis(BaseModel):
    """Data generated by Case Access Project about a Decision."""

    word_count: int
    sha256: str
    orc_confidence: Optional[float] = None
    char_count: int
    pagerank: Optional[PageRank] = None
    cardinality: int
    simhash: str


class DecisionError(Exception):
    """Error for failed attempts to assign Opinions to Decisions."""

    pass


class Decision(BaseModel):
    r"""
    A court decision to resolve a step in litigation.

    Uses the model of a judicial decision from
    the `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.

    One of these records may contain multiple :class:`.Opinion`\s.
    Typically one record will contain all
    the :class:`~justopinion.decisions.Opinion`\s
    from one appeal, but not necessarily from the whole lawsuit. One
    lawsuit may contain multiple appeals or other petitions, and
    if more then one of those generates published Opinions,
    the CAP API will divide those Opinions into a separate
    record for each appeal.
    The outcome of a decision may be determined by one majority
    :class:`~justopinion.decisions.Opinion` or by the combined
    effect of multiple Opinions.
    The lead opinion is commonly, but not always, the only
    Opinion that creates binding legal authority.
    Usually every rule posited by the lead Opinion is
    binding, but some may not be, often because parts of the
    Opinion fail to command a majority of the panel
    of judges.

    :param decision_date:
        date when the opinion was first published by the court
        (not the publication date of the reporter volume)
    :param name:
        full name of the opinion, e.g. "ORACLE AMERICA, INC.,
        Plaintiff-Appellant, v. GOOGLE INC., Defendant-Cross-Appellant"
    :param name_abbreviation:
        shorter name of the opinion, e.g. "Oracle America, Inc. v. Google Inc."
    :param docket_num:
    :param citations:
        ways to cite this Decision
    :param parties:
    :param attorneys:
    :param first_page:
        the page where the opinion begins in its official reporter
    :param last_page:
        the page where the opinion ends in its official reporter
    :param court:
        name of the court that published the opinion
    :param casebody:
        the Decision content including Opinions
    :param juridiction:
        the jurisdiction of the case
    :param cites_to:
        :class:`~justopinion.citations.CAPCitation`\s to other Decisions
    :param id:
        unique ID from CAP API
    :param last_updated:
        date when the record was last updated in CAP API
    :param frontend_url:
        URL to the decision in CAP's frontend
    :param analysis:
        data generated by Case Access Project about this Decision
    """

    decision_date: datetime.date
    name: Optional[str] = None
    name_abbreviation: Optional[str] = None
    docket_num: Optional[str] = None
    citations: Optional[Sequence[CAPCitation]] = None
    parties: List[str] = []
    attorneys: List[str] = []
    first_page: Optional[int] = None
    last_page: Optional[int] = None
    court: Optional[Court] = None
    casebody: Optional[CaseBody] = None
    jurisdiction: Optional[Jurisdiction] = None
    cites_to: Optional[Sequence[CAPCitation]] = None
    id: Optional[int] = None
    last_updated: Optional[datetime.datetime] = None
    frontend_url: Optional[HttpUrl] = None
    analysis: Optional[DecisionAnalysis] = None

    def __str__(self):
        citation = self.citations[0].cite if self.citations else ""
        name = self.name_abbreviation or self.name
        return f"{name}, {citation} ({self.decision_date})"

    @validator("decision_date", pre=True)
    def decision_date_must_include_day(cls, v):
        """Add a day of "01" if a string format is missing it."""
        if isinstance(v, str) and len(v) == 7:
            return v + "-01"
        return v

    @property
    def opinions(self):
        return self.casebody.data.opinions if self.casebody else []

    @property
    def majority(self) -> Optional[Opinion]:
        for opinion in self.opinions:
            if opinion.type == "majority":
                return opinion
        return None

    def find_matching_opinion(
        self,
        opinion_type: str = "",
        opinion_author: str = "",
    ) -> Optional[Opinion]:
        """Get Opinion matching the given attributes."""
        if not opinion_type and not opinion_author:
            if len(self.opinions) >= 1:
                return self.opinions[0]
            return None
        for opinion in self.opinions:
            if ((opinion_type == opinion.type) or not opinion_type) and (
                (opinion_author == opinion.author) or not opinion_author
            ):
                return opinion
        return None

    def add_opinion(self, opinion: Opinion) -> None:
        existing = self.find_matching_opinion(
            opinion_type=opinion.type, opinion_author=opinion.author
        )
        if existing:
            raise DecisionError(
                f"Decision {self} already has an Opinion with type {existing.type} and author {existing.author}."
            )
        self.casebody = self.casebody or CaseBody(data=CaseData())
        self.casebody.data.opinions.append(opinion)



import datetime
import os

from dotenv import load_dotenv
import eyecite
import pytest

from justopinion.citations import CAPCitation
from justopinion.decisions import Decision, Opinion, DecisionError
from justopinion.download import (
    CAPClient,
    CaseAccessProjectAPIError,
    normalize_case_cite,
)


load_dotenv()


class TestDownloads:
    client = CAPClient(api_token=os.getenv("CAP_API_KEY"))

    @pytest.mark.vcr
    def test_download_case_by_id(self):
        case = self.client.fetch(4066790).json()
        assert case["name_abbreviation"] == "Oracle America, Inc. v. Google Inc."
        assert case["id"] == 4066790

    @pytest.mark.default_cassette("TestDownloads.test_download_case_by_id.yaml")
    @pytest.mark.vcr
    def test_read_case_by_id(self):
        case = self.client.read_id(4066790)
        assert case.name_abbreviation == "Oracle America, Inc. v. Google Inc."
        assert case.id == 4066790
        name = "Oracle America, Inc. v. Google Inc., 750 F.3d 1339 (2014-05-09)"
        assert str(case) == name

    @pytest.mark.default_cassette("TestDownloads.test_download_case_by_id.yaml")
    @pytest.mark.vcr
    def test_download_case_by_string_id(self):
        case = self.client.fetch("4066790")
        assert case.json()["name_abbreviation"] == "Oracle America, Inc. v. Google Inc."

    @pytest.mark.vcr
    def test_full_case_by_cite(self):
        response = self.client.fetch("49 F.3d 807", full_case=True)
        lotus = response.json()["results"][0]
        assert lotus["decision_date"] == "1995-03-09"
        assert lotus["casebody"]["data"]["opinions"][0]["author"].startswith("STAHL")

    @pytest.mark.vcr
    def test_read_decision(self):
        decision = self.client.read_cite("1 Breese 34", full_case=True)
        assert isinstance(decision, Decision)
        assert isinstance(decision.opinions[0], Opinion)
        assert decision.opinions[0].author == ""
        assert decision.opinions[0].type == "majority"
        assert str(decision.majority) == "majority opinion"

    @pytest.mark.default_cassette("TestDownloads.test_read_decision.yaml")
    @pytest.mark.vcr
    def test_fetch_decision(self):
        response = self.client.fetch_cite("1 Breese 34", full_case=True)
        assert response.status_code == 200
        result = response.json().get("results")[0]
        assert result["id"] == 435800
        assert result["casebody"]["data"]["head_matter"].startswith("John")

    @pytest.mark.vcr
    def test_fetch_full_decision_by_id(self):
        response = self.client.fetch_id(435800, full_case=True)
        assert response.status_code == 200
        result = response.json()
        assert result["id"] == 435800
        assert result["casebody"]["data"]["head_matter"].startswith("John")

    @pytest.mark.default_cassette("TestDownloads.test_fetch_full_decision_by_id.yaml")
    @pytest.mark.vcr
    def test_fetch_one_decision_and_read(self):
        response = self.client.fetch_id(435800, full_case=True)
        decision = self.client.read_decision_from_response(response)
        assert isinstance(decision, Decision)

    @pytest.mark.default_cassette("TestDownloads.test_full_case_by_cite.yaml")
    @pytest.mark.vcr
    def test_download_and_make_opinion(self):
        cases = self.client.read_decision_list_by_cite(
            cite="49 F.3d 807", full_case=True
        )
        lotus = cases[0]
        lotus_opinion = lotus.majority
        assert lotus_opinion.__class__.__name__ == "Opinion"
        assert str(lotus_opinion) == "majority opinion by STAHL, Circuit Judge"

    @pytest.mark.default_cassette("TestDownloads.test_full_case_by_cite.yaml")
    @pytest.mark.vcr
    def test_repr(self):
        cases = self.client.read_decision_list_by_cite(
            cite="49 F.3d 807", full_case=True
        )
        lotus = cases[0]
        lotus_opinion = lotus.majority
        expected = 'Opinion(type="majority", author="STAHL, Circuit Judge")'
        assert repr(lotus_opinion) == expected

    @pytest.mark.default_cassette("TestDownloads.test_full_case_by_cite.yaml")
    @pytest.mark.vcr
    def test_full_case_as_json(self):
        lotus = self.client.read("49 F.3d 807", full_case=True)
        assert "Lotus" in lotus.name_abbreviation

    def test_error_download_without_case_reference(self):
        with pytest.raises(TypeError):
            self.client.fetch_cite()

    @pytest.mark.vcr
    def test_error_bad_cap_id(self):
        with pytest.raises(CaseAccessProjectAPIError):
            self.client.fetch_id(cap_id=99999999)

    def test_error_bad_cite(self):
        with pytest.raises(ValueError):
            self.client.fetch_cite(cite="999 Cal 9th. 9999")

    @pytest.mark.vcr
    def test_error_full_case_download_without_api_key(self):
        bad_client = CAPClient()
        with pytest.raises(CaseAccessProjectAPIError):
            bad_client.fetch_cite(cite="49 F.3d 807", full_case=True)

    @pytest.mark.vcr
    def test_error_full_case_download_with_wrong_api_key(self):
        bad_client = CAPClient(api_token="aaaaa")
        with pytest.raises(CaseAccessProjectAPIError):
            bad_client.fetch_cite(cite="49 F.3d 807", full_case=True)

    @pytest.mark.vcr
    def test_read_case_from_citation(self):
        citation = CAPCitation(
            cite="460 F. 3d 337",
            reporter="F.3d",
            category="reporters:federal",
            case_ids=[3674823],
            type=None,
        )
        case = self.client.read_cite(citation)
        assert case.id == 3674823

    @pytest.mark.vcr("TestDownloads.test_read_case_from_citation.yaml")
    def test_read_cited_case_from_id_using_client(self):
        case = self.client.read(query=3675682, full_case=False)
        assert case.name_abbreviation == "Kimbrough v. United States"
        cited_case = self.client.read_cite(cite=case.cites_to[0])
        assert cited_case.name_abbreviation == "United States v. Castillo"

    @pytest.mark.vcr
    def test_read_case_list_from_eyecite_case_citation(self):
        case_citation = eyecite.get_citations("9 F. Cas. 50")[0]
        cases_again = self.client.read_decision_list_by_cite(cite=case_citation)
        assert cases_again[0].name_abbreviation == "Fikes v. Bentley"

    @pytest.mark.vcr
    def test_fail_to_read_id_cite(self):
        with pytest.raises(ValueError, match="was type IdCitation, not CaseCitation"):
            self.client.read_decision_list_by_cite(cite="id. at 37")


class TestTextSelection:
    client = CAPClient(api_token=os.getenv("CAP_API_KEY"))

    @pytest.mark.default_cassette("TestDownloads.test_full_case_by_cite.yaml")
    @pytest.mark.vcr
    def test_locate_opinion_text(self):
        lotus = self.client.read("49 F.3d 807", full_case=True)
        opinion = lotus.opinions[0]
        assert opinion.locate_text("method of operation").ranges()[0].start == 12821

    @pytest.mark.default_cassette("TestDownloads.test_full_case_by_cite.yaml")
    @pytest.mark.vcr
    def test_select_opinion_text(self):
        lotus = self.client.read("49 F.3d 807", full_case=True)
        opinion = lotus.opinions[0]
        passage = str(opinion.select_text([(12821, 12840), (12851, 12863)]))
        assert passage == "…method of operation…or procedure…"


class TestDecisions:
    client = CAPClient(api_token="Token " + (os.getenv("CAP_API_KEY") or ""))

    def test_normalize_case_cite(self):
        assert normalize_case_cite("3 US 100") == "3 U.S. 100"

    @pytest.mark.default_cassette("TestDownloads.test_download_case_by_id.yaml")
    @pytest.mark.vcr
    def test_no_majority_without_full_case(self):
        case = self.client.read_id(4066790)
        assert case.name_abbreviation == "Oracle America, Inc. v. Google Inc."
        assert case.majority is None

    @pytest.mark.vcr
    def test_download_case_from_cap_citation(self):
        cite = CAPCitation(
            cite="15 Ill., 284",
            reporter="Ill.",
            category="reporters:state",
            case_ids=[436826],
        )
        case = self.client.read_cite(cite=cite, full_case=True)
        assert case.citations[0].cite == "15 Ill. 284"
        assert str(case.citations[0]) == "Citation to 15 Ill. 284"
        assert str(case.cites_to[0]) == "Citation to 20 Pick. 535"

    @pytest.mark.vcr(
        "TestDownloads.test_read_decision.yaml",
        "TestDecisions.test_download_case_from_cap_citation",
    )
    def test_read_cited_decision(self):
        decision = self.client.read_cite("1 Breese 34", full_case=True)
        assert len(decision.cites_to) == 1
        assert decision.cites_to[0].cite == "15 Ill., 284"
        cited = self.client.read_cite(decision.cites_to[0], full_case=True)
        assert cited.citations[0].cite == "15 Ill. 284"

    @pytest.mark.default_cassette("TestDownloads.test_read_decision.yaml")
    @pytest.mark.vcr
    def test_error_add_opinion_without_specifying_author(self):
        decision = self.client.read_cite("1 Breese 34", full_case=True)
        new_opinion = Opinion(author="")
        with pytest.raises(DecisionError):
            decision.add_opinion(new_opinion)

    @pytest.mark.default_cassette("TestDownloads.test_read_decision.yaml")
    @pytest.mark.vcr
    def test_error_add_opinion_to_blank_decision(self):
        decision = Decision(decision_date=datetime.date(2019, 1, 1))
        new_opinion = Opinion(author="", type="")
        assert len(decision.opinions) == 0
        decision.add_opinion(new_opinion)
        assert len(decision.opinions) == 1

    @pytest.mark.default_cassette("TestDownloads.test_read_decision.yaml")
    @pytest.mark.vcr
    def test_error_add_opinion_with_empty_string_type(self):
        decision = self.client.read_cite("1 Breese 34", full_case=True)
        new_opinion = Opinion(author="", type="")
        with pytest.raises(DecisionError):
            decision.add_opinion(new_opinion)

    @pytest.mark.default_cassette("TestDownloads.test_read_decision.yaml")
    @pytest.mark.vcr
    def test_error_add_same_opinion_twice(self):
        decision = self.client.read_cite("1 Breese 34", full_case=True)
        new_opinion = Opinion(author="Scalia", type="dissent")
        another_opinion = Opinion(author="Scalia", type="dissent")
        decision.add_opinion(new_opinion)
        with pytest.raises(DecisionError):
            decision.add_opinion(another_opinion)



from typing import List, Optional, Union

from eyecite import get_citations
from eyecite.models import CaseCitation
from pydantic import BaseModel


class CAPCitation(BaseModel):
    """
    A case citation generated by the Case Access Project.

    Not to be confused with a CaseCitation created by the Eyecite library.

    :param cite:
        The text making up the citation.
    :param reporter:
        Identifier for the reporter cited to.
    :param category:
        The type of document cited, e.g. "reporters:federal"
    :param case_ids:
        A list of Case Access Project IDs for the cited case.
    :param type:
        The kind of citation, e.g. "official"
    """

    cite: str
    reporter: Optional[str] = None
    category: Optional[str] = None
    case_ids: List[int] = []
    type: Optional[str] = None

    def __str__(self) -> str:
        return f"Citation to {self.cite}"


def normalize_case_cite(cite: Union[str, CaseCitation, CAPCitation]) -> str:
    """Get just the text that identifies a citation."""
    if isinstance(cite, CAPCitation):
        return cite.cite
    if isinstance(cite, str):
        possible_cites = list(get_citations(cite))
        bad_cites = []
        for possible in possible_cites:
            if isinstance(possible, CaseCitation):
                return possible.corrected_citation()
            bad_cites.append(possible)
        error_msg = f"Could not locate a CaseCitation in the text {cite}."
        for bad_cite in bad_cites:
            error_msg += f" {str(bad_cite)} was type {bad_cite.__class__.__name__}, not CaseCitation."
        raise ValueError(error_msg)
    return cite.corrected_citation()



import setuptools

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="justopinion",
    version="0.2.4",
    author="Matt Carey",
    author_email="matt@authorityspoke.com",
    description="Download client for legal opinions",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/mscarey/justopinion",
    project_urls={"Bug Tracker": "https://github.com/mscarey/justopinion/issues"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: Free To Use But Restricted",
        "Natural Language :: English",
    ],
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=["requests", "anchorpoint>=0.5.3", "eyecite>=2.2.0", "pydantic"],
    python_requires=">=3.8",
)



"""Download client for data that can be converted to :class:`~justopinion.decisions.Decision` and :class:`~justopinion.decisions.Opinion` objects."""

from __future__ import annotations

from typing import Dict, List, Optional, Union

import requests

from justopinion.citations import CaseCitation, normalize_case_cite
from justopinion.decisions import CAPCitation, Decision


class CAPClient:
    """Downloads judicial decisions from Case Access Project API."""

    def __init__(self, api_token: Optional[str] = ""):

        """Create download client with an API token and an API address."""
        self.endpoint = "https://api.case.law/v1/cases/"
        if api_token and api_token.startswith("Token "):
            api_token = api_token.split("Token ")[1]
        self.api_token = api_token or ""
        self.api_alert = (
            "To fetch full opinion text using the full_case parameter, "
            "set the CAPClient's 'api_key' attribute to "
            "your API key for the Case Access Project. See https://api.case.law/"
        )

    def fetch(
        self, query: Union[int, str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> requests.models.Response:
        """Query by CAP id or citation, and download Decision from CAP API."""
        if isinstance(query, int) or (isinstance(query, str) and query.isdigit()):
            return self.fetch_id(int(query), full_case=full_case)
        return self.fetch_cite(query, full_case=full_case)

    def read(
        self, query: Union[int, str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> Decision:
        """Query by CAP id or citation, download, and load Decision from CAP API."""
        if isinstance(query, int) or (isinstance(query, str) and query.isdigit()):
            return self.read_id(int(query), full_case=full_case)
        return self.read_cite(query, full_case=full_case)

    def get_api_headers(self, full_case: bool = False) -> Dict[str, str]:
        """Get API headers based on whether the full case text is requested."""
        api_dict = {}
        if full_case:
            if not self.api_token:
                raise CaseAccessProjectAPIError(self.api_alert)
            api_dict["Authorization"] = f"Token {self.api_token}"
        return api_dict

    def fetch_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> requests.models.Response:
        """
        Get the API list response for a queried citation from the CAP API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.
        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.
        :returns:
            the "results" list for this queried citation.
        """

        normalized_cite = normalize_case_cite(cite)

        params = {"cite": normalized_cite}

        headers = self.get_api_headers(full_case=full_case)

        if full_case:
            params["full_case"] = "true"
        response = requests.get(self.endpoint, params=params, headers=headers)
        if response.status_code == 401:
            detail = response.json()["detail"]
            raise CaseAccessProjectAPIError(f"{detail} {self.api_alert}")
        return response

    def read_decision_list_by_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> List[Decision]:
        """
        Download and deserialize the "results" list for a queried citation from the CAP API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.
        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.
        :returns:
            the first case in the "results" list for this queried citation.
        """
        response = self.fetch_cite(cite=cite, full_case=full_case)
        return self.read_decisions_from_response(response)

    def read_decisions_from_response(
        self, response: requests.models.Response
    ) -> List[Decision]:
        """
        Deserialize a list of cases from the "results" list of a response from the CAP API.

        :param response:
            a response from the CAP API

        :returns:
            all decisions in the "results" list for the response
        """
        results = response.json()["results"]
        return [Decision(**result) for result in results]

    def read_decision_from_response(
        self, response: requests.models.Response
    ) -> Decision:
        """
        Deserialize a single case from the "results" list of a response from the CAP API.

        :param response:
            a response from the CAP API

        :returns:
            the first decision in the "results" list for the response
        """
        decision = response.json()
        if "results" in decision:
            return Decision(**decision["results"][0])
        return Decision(**decision)

    def read_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> Decision:
        """
        Download and deserialize a Decision from Caselaw Access Project API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.
        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.
        :returns:
            the first case in the "results" list for this queried citation.
        """
        response = self.fetch_cite(cite=cite, full_case=full_case)
        return self.read_decision_from_response(response=response)

    def fetch_id(
        self, cap_id: int, full_case: bool = False
    ) -> requests.models.Response:
        """
        Download a decision from Caselaw Access Project API.

        :param cap_id:
            an identifier for an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_,
            e.g. 4066790 for
            `Oracle America, Inc. v. Google Inc. <https://api.case.law/v1/cases/4066790/>`_.
        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.
        :returns:
            the first case in the "results" list for this queried citation.
        """
        url = self.endpoint + f"{cap_id}/"
        headers = self.get_api_headers(full_case=full_case)
        params = {}
        if full_case:
            params["full_case"] = "true"
        response = requests.get(url, params=params, headers=headers)
        if cap_id and response.status_code == 404:
            raise CaseAccessProjectAPIError(f"API returned no cases with id {cap_id}")
        return response

    def read_id(self, cap_id: int, full_case: bool = False) -> Decision:
        """
        Download a decision from Caselaw Access Project API.

        :param cap_id:
            an identifier for an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_,
            e.g. 4066790 for
            `Oracle America, Inc. v. Google Inc. <https://api.case.law/v1/cases/4066790/>`_.
        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.
        :returns:
            a Decision created from the first case in the "results" list for
            this queried citation.
        """
        result = self.fetch_id(cap_id=cap_id, full_case=full_case)

        return Decision(**result.json())


class CaseAccessProjectAPIError(Exception):
    """Error for failed attempts to use the Case Access Project API."""

    pass
